package ru.dbotthepony.kstarbound.gl

import org.lwjgl.opengl.GL41
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.api.IStruct3f
import ru.dbotthepony.kstarbound.api.IStruct4F
import ru.dbotthepony.kstarbound.math.FloatMatrix

class ShaderLinkException(reason: String) : RuntimeException(reason)

data class Uniform4f(val x: Float, val y: Float, val z: Float, val w: Float) : IStruct4F
data class Uniform3f(val x: Float, val y: Float, val z: Float) : IStruct3f

class GLUniformLocation(val program: GLShaderProgram, val name: String, val pointer: Int) {
	fun set(value: IStruct4F): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2, v3) = value
		glProgramUniform4f(program.pointer, pointer, v0, v1, v2, v3)
		return this
	}

	fun set(value: IStruct3f): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2) = value
		glProgramUniform3f(program.pointer, pointer, v0, v1, v2)
		return this
	}

	fun set(value: Int): GLUniformLocation {
		program.state.ensureSameThread()
		glProgramUniform1i(program.pointer, pointer, value)
		return this
	}

	fun set(value: FloatMatrix<*>): GLUniformLocation {
		program.state.ensureSameThread()

		if (value.rows == 3 && value.columns == 3) {
			// Матрица 3x3
			glProgramUniformMatrix3fv(program.pointer, pointer, false, value.toFloatArray())
			checkForGLError()
		} else if (value.rows == 4 && value.columns == 4) {
			// Матрица 4x4
			glProgramUniformMatrix4fv(program.pointer, pointer, false, value.toFloatArray())
			checkForGLError()
		} else {
			throw IllegalArgumentException("Can not use matrix with these dimensions: ${value.rows}x${value.columns}")
		}

		return this
	}
}

class GLShaderProgram(val state: GLStateTracker, vararg shaders: GLShader) {
	val pointer = glCreateProgram()
	var linked = false
		private set

	private val attached = HashSet<GLShader>()
	val access = object : Collection<GLShader> by attached {}

	operator fun get(name: String): GLUniformLocation? {
		state.ensureSameThread()
		check(linked) { "Shader program is not linked!" }

		val location = glGetUniformLocation(pointer, name)

		if (location == -1)
			return null

		return GLUniformLocation(this, name, location)
	}

	operator fun set(name: String, value: Uniform4f) = this[name]?.set(value)
	operator fun set(name: String, value: Uniform3f) = this[name]?.set(value)
	operator fun set(name: String, value: Int) = this[name]?.set(value)
	operator fun set(name: String, value: FloatMatrix<*>) = this[name]?.set(value)

	fun attach(shader: GLShader) {
		state.ensureSameThread()
		check(!linked) { "Already linked!" }

		if (!attached.add(shader)) {
			throw IllegalStateException("Already attached! $shader")
		}

		glAttachShader(pointer, shader.pointer)
	}

	fun link() {
		check(!linked) { "Already linked!" }
		glLinkProgram(pointer)

		val success = intArrayOf(0)
		glGetProgramiv(pointer, GL_LINK_STATUS, success)

		if (success[0] == 0) {
			throw ShaderLinkException(glGetShaderInfoLog(pointer))
		}

		glGetError()

		linked = true
	}

	fun use() = state.use(this)

	fun unlinkChildren() {
		attached.forEach(GLShader::unlink)
	}

	init {
		shaders.forEach(this::attach)
	}
}
